<?php

namespace Drupal\vertex_ai_search\Service;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\search\Entity\SearchPage;
use Drupal\search\SearchPageRepositoryInterface;
use Drupal\vertex_ai_search\VertexAutocompletePluginManager;
use Drupal\vertex_ai_search\VertexSearchFilterPluginManager;
use Drupal\vertex_ai_search\VertexSearchResultsPluginManager;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Google\ApiCore\ApiException;
use Google\Cloud\DiscoveryEngine\V1\Client\SearchServiceClient;
use Google\Cloud\DiscoveryEngine\V1\SearchRequest;
use Google\Cloud\DiscoveryEngine\V1\SearchRequest\SpellCorrectionSpec;
use Google\Cloud\DiscoveryEngine\V1\SearchRequest\SpellCorrectionSpec\Mode;

/**
 * Gets Dynamic List Manager.
 */
class VertexSearchManager implements VertexSearchManagerInterface {

  use StringTranslationTrait;

  /**
   * Search Page Repository Manager.
   *
   * @var \Drupal\search\SearchPageRepositoryInterface
   */
  protected $searchPageRepository;

  /**
   * Vertex Autocomplete Plugin Manager.
   *
   * @var \Drupal\vertex_ai_search\VertexAutocompletePluginManager
   */
  protected $autocompletePluginManager;

  /**
   * Vertex Search Filter Plugin Manager.
   *
   * @var \Drupal\vertex_ai_search\VertexSearchFilterPluginManager
   */
  protected $searchFilterPluginManager;

  /**
   * Vertex Search Results Plugin Manager.
   *
   * @var \Drupal\vertex_ai_search\VertexSearchResultsPluginManager
   */
  protected $searchResultsPluginManager;

  /**
   * Object constructor.
   *
   * @param \Drupal\search\SearchPageRepositoryInterface $search_page_repository
   *   Search page repository manager.
   * @param \Drupal\vertex_ai_search\VertexAutocompletePluginManager $autocomplete_plugin_manager
   *   Vertex AI Search Autocomplete Manager.
   * @param \Drupal\vertex_ai_search\VertexSearchFilterPluginManager $search_filter_plugin_manager
   *   Vertex AI Search filter plugin manager.
   * @param \Drupal\vertex_ai_search\VertexSearchResultsPluginManager $search_results_plugin_manager
   *   Vertex AI Search results plugin manager.
   */
  public function __construct(
    SearchPageRepositoryInterface $search_page_repository,
    VertexAutocompletePluginManager $autocomplete_plugin_manager,
    VertexSearchFilterPluginManager $search_filter_plugin_manager,
    VertexSearchResultsPluginManager $search_results_plugin_manager,
  ) {
    $this->searchPageRepository = $search_page_repository;
    $this->autocompletePluginManager = $autocomplete_plugin_manager;
    $this->searchFilterPluginManager = $search_filter_plugin_manager;
    $this->searchResultsPluginManager = $search_results_plugin_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function executeSearch(string|Object $search_page, array $search_parameters) {

    // Retrieve configuration of specified search page (or search page object).
    $searchPageConfig = $this->retrieveSearchPageConfig($search_page);

    // Prepare a request to send to Vertex AI Search.
    $searchRequest = $this->prepareSearchRequest(
      $searchPageConfig,
      $search_parameters
    );

    // Perform Vertex Search.
    $results = $this->performVertexSearch($searchPageConfig, $searchRequest);

    // Retrieve curated results from VertexSearchResults plugin instances.
    $results['curated'] = $this->retrievePluginCuratedResults($searchPageConfig, $search_parameters);

    return $results;

  }

  /**
   * Helper to retrieve the configuration of the relevant custom search page.
   *
   * @param string|Object $search_page
   *   Machine name of custom search page or an object representation of page.
   *
   * @return array
   *   An array of custom search page configuration.
   */
  private function retrieveSearchPageConfig(string|Object $search_page) {

    // Goal is to get the configuration of the SearchPage.
    $searchPageConfig = [];

    // If $search_page is string, retrieve config from active pages array.
    if (is_string($search_page)) {
      $searchPages = $this->searchPageRepository->getActiveSearchPages();
      $searchPageConfig = isset($searchPages[$search_page]) ?
        $searchPages[$search_page]->get('configuration') : [];
    }
    elseif (($search_page instanceof SearchPage) && $search_page->get('status') === TRUE) {
      $searchPageConfig = $search_page->get('configuration');
    }

    return $searchPageConfig;

  }

  /**
   * Helper to retrieve the configuration of the relevant custom search page.
   *
   * @param array $search_page_config
   *   Configuration of the relevant custom search page.
   * @param array $search_parameters
   *   Array of search parameters (keys, page, etc...).
   *
   * @return \Google\Cloud\DiscoveryEngine\V1\SearchRequest
   *   A Vertex Search Request ready to be sent.
   */
  protected function prepareSearchRequest(array $search_page_config, array $search_parameters) {

    // Configure Search Client with serving configuration.
    // @see: https://cloud.google.com/php/docs/reference/cloud-discoveryengine/0.4.0/V1.Client.SearchServiceClient.
    $formattedServingConfig = SearchServiceClient::servingConfigName(
      $search_page_config['google_cloud_project_id'],
      $search_page_config['google_cloud_location'],
      $search_page_config['vertex_ai_data_store_id'],
      $search_page_config['vertex_ai_serving_config']
    );

    // Prepare the search request.
    $request = (new SearchRequest())->setServingConfig($formattedServingConfig);

    // Set Keywords to be used in request query.
    $request->setQuery($search_parameters['keys']);

    // Set results per page.
    $request->setPageSize($search_page_config['resultsPerPage']);

    // Specify if safe search is on or off.
    $safeSearch = !empty($search_page_config['safeSearch']) ? TRUE : FALSE;
    $request->setSafeSearch($safeSearch);

    // Specify the spelling correction mode (automatic or not).
    $spellCorrection = new SpellCorrectionSpec();
    $spellCorrection->setMode(MODE::value($search_page_config['spelling_correction_mode']));
    $request->setSpellCorrectionSpec($spellCorrection);

    // Set offset (starting point) of search request.
    $page = $search_page_parameters['page'] ?? 0;
    $offset = $search_page_config['resultsPerPage'] * $page;
    $request->setOffset($offset);

    // Set filter for search request if configured.
    $filter = $this->retrievePluginFilter($search_page_config);
    if ($filter) {
      $request->setFilter($filter);
    }

    return $request;

  }

  /**
   * Helper to perform a Vertex Search.
   *
   * @param array $search_page_config
   *   Configuration of the relevant custom search page.
   * @param \Google\Cloud\DiscoveryEngine\V1\SearchRequest $search_request
   *   A Vertex Search Request ready to be sent.
   *
   * @return array
   *   An array containing the results of the search.
   */
  protected function performVertexSearch(array $search_page_config, SearchRequest $search_request) {

    $results = [];

    // Retrieve credentials and create a search client.
    // @see https://github.com/googleapis/google-cloud-php/blob/main/AUTHENTICATION.md.
    $credPath = $search_page_config['service_account_credentials_file'];
    $searchServiceClient = new SearchServiceClient([
      'credentials' => json_decode(file_get_contents($credPath), TRUE),
    ]);

    // Call the API and handle any network failures.
    try {

      /** @var \Google\ApiCore\PagedListResponse $response */
      $response = $searchServiceClient->search($search_request);

      $page = $response->getPage();

      $responseObject = $page->getResponseObject();
      $results['totalResults'] = $responseObject->getTotalSize();
      $results['pageResultCount'] = $page->getPageElementCount();
      $results['correctedQuery'] = $responseObject->getCorrectedQuery();

      // Determine query correction status.
      $results['queryCorrected'] = FALSE;
      $correctConfig = $search_page_config['spelling_correction_mode'];
      if (!empty($results['correctedQuery']) &&
         ($correctConfig === 'AUTO' || $correctConfig === 'MODE_UNSPECIFIED')) {
        $results['queryCorrected'] = TRUE;
      }

      /** @var \Google\Cloud\DiscoveryEngine\V1\SearchResponse\SearchResult $result */
      foreach ($page as $result) {
        $results['results'][] = json_decode($result->serializeToJsonString(), TRUE);
      }

    }
    catch (ApiException $ex) {
      printf('Call failed with message: %s' . PHP_EOL, $ex->getMessage());
    }

    // Determine which configuration items to send to user.
    $results['search_page_config'] = $this->retrieveResponseConfig($search_page_config);

    return $results;

  }

  /**
   * Helper function to prep config array to be included in response.
   *
   * This is configuration information that may be helpful to the client.
   *
   * @param array $search_page_config
   *   The full configuration array.
   *
   * @return array
   *   The configuration to be included in the response.
   */
  protected function retrieveResponseConfig(array $search_page_config) {

    $responseConfig = [
      'autocomplete_enable',
      'autocomplete_trigger_length',
      'autocomplete_max_suggestions',
      'resultsPerPage',
      'totalResultsLimit',
      'result_parts',
      'removeDomain',
      'results_message',
      'results_message_singular',
      'no_results_message',
      'no_keywords_message',
    ];

    $responseConfiguration = array_intersect_key($search_page_config, array_flip($responseConfig));

    return $responseConfiguration;

  }

  /**
   * Gets the formatted filter from the selected plugin.
   *
   * @param array $search_page_config
   *   Array containing configuration of relevant custom search page.
   *
   * @return false|string
   *   The formatted filter or FALSE.
   */
  private function retrievePluginFilter(array $search_page_config) {

    if (empty($search_page_config['filter_enable']) || empty($search_page_config['filter_plugin'])) {
      return FALSE;
    }

    $filterPlugin = NULL;
    try {
      $filterPlugin = $this->searchFilterPluginManager->createInstance(
        $search_page_config['filter_plugin'],
        $search_page_config
      );
    }
    catch (PluginException $e) {
      printf('Unable to create filter plugin instance: %s', $e->getMessage());
    }

    return $filterPlugin->getSearchFilter();

  }

  /**
   * Gets results manually curated by search results plugins.
   *
   * @param array $search_page_config
   *   Array containing configuration of relevant custom search page.
   * @param array $search_page_parameters
   *   Array containing parameters of the search.
   *
   * @return array
   *   An array of curated items to include in the search.
   */
  private function retrievePluginCuratedResults(array $search_page_config, array $search_page_parameters) {

    // Curated Results.
    $curated = [];

    // Retrieve any VertexSearchResultPlugins and curated results.
    $resultsPluginDefinitions = $this->searchResultsPluginManager->getDefinitions();

    foreach ($resultsPluginDefinitions as $pluginKey => $pluginDefinition) {

      $resultsPlugin = $this->searchResultsPluginManager->createInstance(
        $pluginKey
      );

      $curated = $resultsPlugin->retrieveCuratedResults($search_page_config, $search_page_parameters);

    }

    return $curated;
  }

}
